%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                         Code for EE367 Project
%             Simulation of Light field Fundus Photography
%                      Sha Tong & T. J Melanson
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Note                        
% *This script simulates light field camera image formation through ray tracing            
% *The script takes long time to run due to large amount of rays traced
% *Typical computational time is about 1~2 hours on a power mac
% *Final size of the matrix storing lightfield data is more than 1.2G
% *Three major parts in this code:
% (1) establish the objective as the input to lightfield camera
% (2) ray tracing in lightfield camera, data stored in a 5D matrix
% (3) data processing and plotting  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part A: establish the objective                       
% Assign pixel value to a 3D retina with defined FOV
% Each pixel on retina under simulation is imaged by eye lens at near inf
% The relationship between objective on retina and its exit angle from eye
% is modeled in Zemax using Gullstrand eye objective model: 
% See details: http://www.zemax.com/os/resources/learn/knowledgebase/zemax-models-of-the-human-eye
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close all;
clear all;
% Read objective
I=im2double(imread('Fundus.jpg'));
% Increase the density of pixel for ray tracing
for n=1:3
  a=I(:,:,n);
  a=interp2(a,3);
  obj(:,:,n)=a;
end
% get size
siz1=size(obj,1);
siz2=size(obj,2);
siz3=size(obj,3);
% Define the corresponding angle of view for the objective
FOV=30; 
% Diameter of the eye ball in mm
Eye_D=24;
% Corresponding physical size of the objective
obj_s=(FOV*2/360)*pi*Eye_D;
% pupil size
pupil=4;
% Load the beam exit angle vs raidal angle on retina
ZMX_ang=dlmread('Eye angle.txt', '', 8, 0);
% Establish the incidence angle into imaging lens
% Each angle representing a ray bundle of size 4mm
Retina_ang_x=linspace(-FOV/2+0.5*FOV/siz1, FOV/2-0.5*FOV/siz1, siz1);
Retina_ang_y=linspace(-FOV/2+0.5*FOV/siz2, FOV/2-0.5*FOV/siz2, siz2);
Retina_ang_x_exit=interp1(ZMX_ang(:,1),ZMX_ang(:,2), Retina_ang_x);
Retina_ang_y_exit=interp1(ZMX_ang(:,1),ZMX_ang(:,2), Retina_ang_y);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part B: ray tracing to eastablish the light field                       
% Imaging lens is at f away from the pupil
% Telescope can be place infornt of the imaging lens to readue the size of
% objective (0.7X used in this simulation)
% Microlens array located at f away from the imaging lens
% Sensor array located at micro_f away from the micrlens array
% Each ray is characterized by 4DOF (x,y, ceta, beta)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% size of image lens=size of lens array=size of sensor array
% image lens aperature 
AP=12; 
% image lens focal length 
f=24;
% Field curvature for single lens, ~ equal to the focal length
Rf=f;
% imaging xy location 
ns=tan(repmat(Retina_ang_x_exit,siz2,1)/57.3)*0.7*f;
nv=tan(repmat(Retina_ang_y_exit',1,siz1)/57.3)*0.7*f;
% imaging depth location
vmatrix=sqrt(Rf^2.-ns.^2-nv.^2);
% microlens focal
micro_f=0.04;
% micro lens size
micro_d=0.02;
% #s of micro lens per row
micro_N=AP/micro_d;
% sensor pixel size diameter in mm
sen_d=0.001; 
% # of sensor per row
sen_N=micro_N*micro_d/sen_d;
% # of sensor after each micro_lens /row
sen_N_lens=micro_d/sen_d;
% Sensor total size in mm
sen_s=sen_d*sen_N;
% total assigned ray per pixel
TN_rays=1000;
% #s of rays in each direction (x or y)
N_rays=round(sqrt(TN_rays));  
% array location
array_loc=max(vmatrix(:)); 
% sensor location 
sen_loc=array_loc+micro_f;  
% initial lightField
lightField=zeros(micro_N, micro_N,sen_N_lens, sen_N_lens, siz3);
% per color
for l=1:siz3
    % Per x pixel
    for n=1:siz1 
    % per y pixel
        for m=1:siz2   
            % discretize lens in x direction
            nX=linspace(ns(n,m)-pupil/2,ns(n,m)+pupil/2,N_rays);
            % discretize lens in y direction
            nY=linspace(nv(n,m)-pupil/2,nv(n,m)+pupil/2,N_rays);
            % calcuate ceta
            ceta=atan((ns(n,m)-nX)/vmatrix(n,m));  
            % calcuate beta
            beta=atan((nv(n,m)-nY)/vmatrix(n,m));  
            % x location of ray landing on array plane
            array_x=nX+tan(ceta)*array_loc;     
            % y location of ray landing on array plane
            array_y=nY+tan(beta)*array_loc; 
            % register sensor location of single ray
            for i=1:N_rays
                for j=1:N_rays
                    % array x pixel value
                    array_x_pix=ceil((array_x(i)+0.5*micro_N*micro_d)/micro_d);
                    % array y pixel value
                    array_y_pix=ceil((array_y(j)+0.5*micro_N*micro_d)/micro_d);  
                    if((array_x_pix<=micro_N) && (array_y_pix<=micro_N))  
                        % find the center of pixel to transfer the
                        % coordinate from global to local 
                        array_x_pix_cen=(array_x_pix-1)*micro_d+0.5*micro_d-0.5*micro_N*micro_d;
                        array_y_pix_cen=(array_y_pix-1)*micro_d+0.5*micro_d-0.5*micro_N*micro_d;
                        if ((array_x(i)-array_x_pix_cen)^2+(array_y(j)-array_y_pix_cen)^2<=2*(micro_d/2)^2)
                            % xy location of incident ray on micro lens
                            x_in=array_x(i)-array_x_pix_cen;
                            y_in=array_y(j)-array_y_pix_cen;
                            % xy location of the ray on sensor
                            x_out=tan(-x_in/micro_f+ceta(i))*micro_f+x_in;
                            y_out=tan(-y_in/micro_f+beta(j))*micro_f+y_in;
                            % angle of the ray reaching sensor
                            ceta_out=ceta(i)-x_in/micro_f;
                            beta_out=beta(j)-y_in/micro_f; 
                            % sensor xy value, global
                            sen_x=array_x_pix_cen+x_out;
                            sen_y=array_y_pix_cen+y_out;
                            if ((abs(sen_x)<=0.5*sen_s) && (abs(sen_y)<=0.5*sen_s))
                                % sensor xy pixel #
                                sen_x_pix=ceil((x_out+0.5*micro_d)/sen_d);
                                sen_y_pix=ceil((y_out+0.5*micro_d)/sen_d);
                                if (sen_x_pix~=0) && (sen_x_pix<=sen_N_lens) &&(sen_x_pix>0) && (sen_y_pix~=0) && (sen_y_pix<=sen_N_lens) &&(sen_y_pix>0)
                                    % read out the sensor pixel value
                                    mid_pix=lightField(array_x_pix,array_y_pix, sen_x_pix,sen_y_pix,l);  
                                    % add new ray
                                    mid_pix=mid_pix+(obj(n,m,l)/N_rays^2)*sqrt(cos(ceta_out)^2+cos(beta_out)^2);  
                                    % new pixel reading
                                    % register the micro lens x, y and
                                    % sensor x, y for this ray
                                    lightField(array_x_pix,array_y_pix,sen_x_pix,sen_y_pix,l)=mid_pix;    
                                end
                            end
                        end
                    end
                end
            end
        end
     end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part C: Image processing and plotting                       
% Use lightfiled matrix to generate focal stack, all in focus image, different viewing perspective and depth map 
% Focal stack in Imgseries matrix through shift and add
% All focus image based on max gradient approach
% Different viewing persepctive by picking corresponding pixels after each
% micro lens
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Average image by sum all fields 
sz1=micro_N*sen_N_lens;
sz2=sz1;
sz3=siz3;
microlensSize=sen_N_lens;
sum_img=zeros(sz1/microlensSize, sz2/microlensSize,sz3);
% Adding each angle
for kx=1:microlensSize
  for ky=1:microlensSize
     sum_img=sum_img+squeeze(lightField(:,:,kx,ky,:)); 
  end
end
figure;
imshow(sum_img/max(sum_img(:))); title('Sum image');

% plot different viewing perspective
figure;
view=squeeze((lightField(:,:,10,10,:))); 
subplot(3,3,5);imshow(view/max(view(:)));title('center view');
view=squeeze((lightField(:,:,15,10,:)));  
subplot(3,3,2);imshow(view/max(view(:)));title('top view');
view=squeeze((lightField(:,:,10,15,:)));
subplot(3,3,4);imshow(view/max(view(:)));title('left view');
view=squeeze((lightField(:,:,10,6,:)));
subplot(3,3,6);imshow(view/max(view(:))); title('right view');
view=squeeze((lightField(:,:,6,10,:)));
subplot(3,3,8);imshow(view/max(view(:))); title('bottom view');

% Shift and add to generate focal stack
% can set scaling factor high if need to shift sub pixel
scalingf=1;  
% matrix extension to accomodate shift
extend=100;  
 % reserve 50 pixel to four edges to allow shift, so the maximum shift allowed is 50 pixels
xstarting=round(extend/2); 
% xending-xstarting+1=extended matrix elements
xending=xstarting+(scalingf*sz1/microlensSize)-1;  
ystarting=xstarting;
yending=ystarting+(scalingf*sz2/microlensSize)-1;
% max angle
maxU=(microlensSize-1)/2; 
% 10 step to cover focal stack, can adjust 
focal_stack_sz=10;
% Imgseries to store focal stack
Imgseries=zeros(scalingf*sz1/microlensSize+extend,scalingf*sz2/microlensSize+extend,sz3,focal_stack_sz);
for m=1:10
    sum_img=zeros(scalingf*sz1/microlensSize+extend,scalingf*sz2/microlensSize+extend,3); 
    I=sum_img;
    % since sensor at max(v), only need to shift to one direction
    maxshift=m-1;
    for kx=1:microlensSize
        % U coordinates
        u=(kx-1-maxU); 
        % U shift
        ushift=round((u/maxU)*maxshift*scalingf);  
        for ky=1:microlensSize  
            v=(ky-1-maxU);   
            vshift=round((v/maxU)*maxshift*scalingf);
            % perform shift for each color
            for n=1:sz3
                % each image with the same U, V and color
                temp_img=squeeze(lightField(:,:,kx,ky,n));   
                % shift the image by store it into I matrix with new coordinates
                I(xstarting-ushift:xending-ushift, ystarting-vshift:yending-vshift,n)=temp_img;  
            end
            % sum the shifted image
            sum_img=sum_img+I;  
            % reset I
            I=zeros(scalingf*sz1/microlensSize+extend,scalingf*sz2/microlensSize+extend,3);  
        end
    end
    Imgseries(:,:,:,m)=sum_img;
end

% Gradient calculation 
[sz1,sz2,sz3,sz4]=size(Imgseries);  
AGmag=zeros(sz1,sz2,sz4); 
% for every shifted image
for m=1:sz4
    % read image from Imgseries
    img=squeeze(Imgseries(:,:,:,m));  
    SGmag=zeros(sz1,sz2);  % reset
    for n=1:sz3  % 3 color chanel
        % get gradient
        [Gmag,Gdir] = imgradient(img(:,:,n),'intermediate'); 
        % sum up color chanel
        SGmag=SGmag+Gmag;  
    end
    % store the gradient for image to AGmag matrix
    AGmag(:,:,m)=SGmag;   
end

% Max gradient for the same pixel across 
% I matrix stores the image #s corresponding to the max gradient for every pixel
[M,I]=max(AGmag, [], 3);

% Generate all focus img 
Allfocusimg=zeros(sz1,sz2,sz3);
% search across the shifted images
for m=1:sz4
  img=zeros(sz1,sz2);
  for n=1:3   % search all color chanels
    img=squeeze(Imgseries(:,:,n,m));  % img with certain color
    % img*(I==m) pick the pixels from the image has the largest gradient
    Allfocusimg(:,:,n)=Allfocusimg(:,:,n)+img.*(I==m);  
  end
end
Figure;
imshow(Allfocusimg/max(Allfocusimg(:))); title('all focus image');

% Depth map
depthmap=zeros(sz1,sz2);
for m=1:sz4
  depthmap=depthmap+(I==m)*(m-6)*0.9;
end

% HSV depth map
normald=depthmap-min(depthmap(:));
% H is the normalized depth (shift)
H=normald/max(normald(:));  
% S is constant 1
S=ones(sz1,sz2);    
normalv=M-min(M(:));
% V is normlized gradient together with H
V=normalv/max(normalv(:));  
img=cat(3,H,S,V);
rgbimg=hsv2rgb(img);
figure; 
% adjust scaling for better view
imagesc(rgbimg.^(1/2.2));
title('depth map');


