%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                         Code for EE367 Project
%             Simulation of Normal Fundus Photography
%                      Sha Tong & T. J Melanson
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Note                        
% *This script simulates normal camera image formation through ray tracing            
% *The script takes long time to run due to large amount of rays traced
% *Typical computational time is about half hour on a power mac
% *Final size of the matrix storing lightfield data is about 0.5G
% *Three major parts in this code:
% (1) establish the objective as the input to lightfield camera
% (2) ray tracing in lightfield camera, data stored in a 5D matrix
% (3) data processing and plotting  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part A: establish the objective                       
% Assign pixel value to a 3D retina with defined FOV
% Each pixel on retina under simulation is imaged by eye lens at near inf
% The relationship between objective on retina and its exit angle from eye
% is modeled in Zemax using Gullstrand eye objective model: 
% See details: http://www.zemax.com/os/resources/learn/knowledgebase/zemax-models-of-the-human-eye
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Read objective
I=im2double(imread('Fundus.jpg'));
% Increase the density of pixel for ray tracing
for n=1:3
  a=I(:,:,n);
  a=interp2(a,3);
  obj(:,:,n)=a;
end
% get size
siz1=size(obj,1);
siz2=size(obj,2);
siz3=size(obj,3);
% Define the corresponding angle of view for the objective
FOV=30; 
% Diameter of the eye ball in mm
Eye_D=24;
% Corresponding physical size of the objective
obj_s=(FOV*2/360)*pi*Eye_D;
% pupil size
pupil=4;
% Load the beam exit angle vs raidal angle on retina
ZMX_ang=dlmread('Eye angle.txt', '', 8, 0);
% Establish the incidence angle into imaging lens
% Each angle representing a ray bundle of size 4mm
Retina_ang_x=linspace(-FOV/2+0.5*FOV/siz1, FOV/2-0.5*FOV/siz1, siz1);
Retina_ang_y=linspace(-FOV/2+0.5*FOV/siz2, FOV/2-0.5*FOV/siz2, siz2);
Retina_ang_x_exit=interp1(ZMX_ang(:,1),ZMX_ang(:,2), Retina_ang_x);
Retina_ang_y_exit=interp1(ZMX_ang(:,1),ZMX_ang(:,2), Retina_ang_y);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part B: ray tracing to eastablish the light field                       
% Imaging lens is at f away from the pupil
% Telescope can be place infornt of the imaging lens to readue the size of
% objective (0.7X used in this simulation)
% Microlens array located at f away from the imaging lens
% Sensor array located at micro_f away from the micrlens array
% Each ray is characterized by 4DOF (x,y, ceta, beta)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% size of image lens=size of lens array=size of sensor array
% image lens aperature 
AP=12; 
% image lens focal length 
f=24;
% Field curvature for single lens, ~ equal to the focal length
Rf=0.8*f;
% Magnification
Mag=0.7;
% imaging xy location 
ns=tan(repmat(Retina_ang_x_exit,siz2,1)/57.3)*Mag*f;
nv=tan(repmat(Retina_ang_y_exit',1,siz1)/57.3)*Mag*f;
% imaging depth location
vmatrix=sqrt(Rf^2.-ns.^2-nv.^2);

% size of the sensor
sen_d=2*(ns(ceil(siz1/2+1),ceil(siz1/2+1))-ns(ceil(siz1/2),ceil(siz1/2))); 
% # of sensor/row
sen_N=ceil(AP/sen_d);
% Sensor total size in mm
% Sensor has the same size as AP
sen_s=AP;
% Total #s of ray to simulate
TN_rays=1000;
% Ray per direction
N_rays=round(sqrt(TN_rays));  
% image point wise locations
sen_loc=max(vmatrix(:));  
% Initial values
image=zeros(sen_N, sen_N,siz3);
% color
for l=1:siz3  
    % x direction pixel
    for n=1:siz1   
        % y direction pixel
        for m=1:siz2  
            % X direction lens discretize
            nX=linspace(ns(n,m)-pupil/2,ns(n,m)+pupil/2,N_rays);
            % X direction lens discretize
            nY=linspace(nv(n,m)-pupil/2,nv(n,m)+pupil/2,N_rays);
            % Calcuate ceta angle
            ceta=atan((ns(n,m)-nX)/vmatrix(n,m));  
            % Calcuate beta angle
            beta=atan((nv(n,m)-nY)/vmatrix(n,m));  
            % Sensor location x
            sen_x=nX+tan(ceta)*sen_loc;   
            % Sensor location y
            sen_y=nY+tan(beta)*sen_loc;   
            % For every single ray
            for i=1:N_rays
                for j=1:N_rays
                    % Round lens, ray need to land on sensor
                    if (nX(i)^2+nY(j)^2<=(AP/2)^2) && ((abs(sen_x(i))<=0.5*sen_s) && (abs(sen_y(j))<=0.5*sen_s))
                        % sen X pixel value
                        sen_x_pix=ceil((sen_x(i)+0.5*sen_N*sen_d)/sen_d);
                        % sen Y pixel value
                        sen_y_pix=ceil((sen_y(j)+0.5*sen_N*sen_d)/sen_d);  
                        if (sen_x_pix~=0) && (sen_x_pix<=sen_N) &&(sen_x_pix>0) && (sen_y_pix~=0) && (sen_y_pix<=sen_N) &&(sen_y_pix>0)
                            mid_pix=image(sen_x_pix,sen_y_pix,l);  % previous pixel reading
                            mid_pix=mid_pix+(obj(n,m,l)/N_rays^2);  % adding new ray
                            image(sen_x_pix,sen_y_pix,l)=mid_pix;  % new pixel reading  
                        end
                    end
                end
            end
        end
    end
end
imshow(image/max(image(:))); title('Fundus image');














